from enum import Enum
from pathlib import Path
from typing import List, Optional

from pandas import DataFrame, ExcelFile

from processing.config.ExcelSheet import ExcelSheet


class ConfigParser:
    def __init__(
        self,
        path: str,
        folder: str,
    ) -> None:
        self.path = self.validate(path)
        self.folder = folder
        self.excel = ExcelFile(self.path)

    def validate(self, path: str) -> str:
        assert path is not None, "Please define a path"
        assert Path(path).exists(), f"Path {path} does not exist"
        return path

    def parse(self, sheet: ExcelSheet) -> DataFrame:
        return self.excel.parse(sheet.value)  # type: ignore

    def get(
        self,
        sheet_name: ExcelSheet,
        column: Enum,
        suffix: Optional[str] = None,
    ) -> List:
        if suffix is not None:
            path = f"{column.value}{suffix}"
        else:
            path = column.value

        sheet = self.parse(sheet_name)
        series = sheet[path]
        list_ = [value for value in series]
        return list_



from typing import List, Tuple

from processing.utils.validate_int import validate_int


class RangeConfig:
    index: int
    name: str
    start: int
    end: int

    def __init__(
        self,
        index: int,
        name: str,
        start: int,
        end: int,
    ) -> None:
        validate_int(start)
        validate_int(end)

        self.index = index
        self.name = name
        self.start = start
        self.end = end

    @staticmethod
    def flatten(
        ranges: List["RangeConfig"],
    ) -> Tuple[List[str], List[int], List[int]]:
        names = [r.name for r in ranges]
        starts = [r.start for r in ranges]
        ends = [r.end for r in ranges]

        return names, starts, ends

    @staticmethod
    def reconstruct(
        names: List[str],
        starts: List[int],
        ends: List[int],
    ) -> List["RangeConfig"]:
        ranges = []

        for index, name in enumerate(names):
            range_ = RangeConfig(
                index=index,
                name=name,
                start=starts[index],
                end=ends[index],
            )

            ranges.append(range_)

        return ranges



import numpy


def validate_int(value: int) -> None:
    if isinstance(value, int):
        return

    if type(value) is numpy.int64 or numpy.int32:
        return

    raise TypeError(f"Unable to validate int for {value}")



from typing import Dict, List, Literal, Optional

import numpy as np
import pydub
from pydub import AudioSegment

from processing.config.bands.BandConfig import BandConfig
from processing.config.files.FileConfig import FileConfig
from processing.loaders.SoundSlice import SoundSlice
from processing.loaders.Spectrogram import Spectrogram

FilteredKey = str
SoundSliceStartEnd = str
SpectrogramMode = Literal["psd", "amplitude"]
Spectrograms = Dict[SpectrogramMode, Dict[SoundSliceStartEnd, Spectrogram]]


class SoundLoader:
    __audio: Optional[AudioSegment] = None
    __sample_rate: Optional[int] = None

    @property
    def audio(self) -> AudioSegment:
        assert self.__audio is not None, "Please load file"
        return self.__audio

    @audio.setter
    def audio(self, audio: AudioSegment) -> None:
        self.__audio = audio

    @property
    def sample_rate(self) -> int:
        assert self.__sample_rate is not None, "Please load file"
        return self.__sample_rate

    @sample_rate.setter
    def sample_rate(self, sample_rate: int) -> None:
        self.__sample_rate = sample_rate

    def destroy(self):
        # this will raise `AttributeError` if you try loading the file again
        del self.__audio
        del self.__sample_rate
        del self.__filtered
        del self.__spectrograms

    def release(self):
        self.__audio = None
        self.__sample_rate = None
        self.__initialize_filtered()
        self.__initialize_spectrogram()

    def print_leftovers(self):
        print("Audio: ", type(self.__audio))
        print("Sample rate: ", type(self.__sample_rate))
        print("Filtered keys: ", self.__filtered.keys())
        print("Spectrogram psd: ", self.__spectrograms["psd"].keys())
        print("Spectrogram amplitude: ", self.__spectrograms["amplitude"].keys())

    def __initialize_filtered(self):
        self.__filtered: Dict[FilteredKey, List[float]] = {}

    def __initialize_spectrogram(self):
        self.__spectrograms: Spectrograms = {
            "psd": {},
            "amplitude": {},
        }

    def load(self, file: FileConfig):
        # print(f"SoundLoader: file index {file.index}")
        sound: pydub.AudioSegment = AudioSegment.from_file(file.path)
        sample_rate: int = sound.frame_rate  # type: ignore

        self.audio = sound
        self.sample_rate = sample_rate

        self.__initialize_filtered()
        self.__initialize_spectrogram()

        return self.audio, sample_rate

    def slice(
        self,
        start: int,
        end: int,
    ) -> AudioSegment:
        assert self.audio is not None, "not loaded"
        slice_: AudioSegment = self.audio[start:end]  # type: ignore
        return slice_

    def get_filtered(self, band: BandConfig):
        from maad import sound

        kf = f"{band.low}-{band.high}"

        if kf in self.__filtered.keys():
            return self.__filtered[kf]

        samples = self.audio.get_array_of_samples()

        filtered: np.ndarray = sound.select_bandwidth(
            x=samples,
            fs=self.sample_rate,
            fcut=[band.low, band.high],
            forder=6,
            fname="butter",
            ftype="bandpass",
        )

        self.__filtered[kf] = filtered.tolist()
        return self.__filtered[kf]

    def get_spectrogram(
        self,
        slice_: SoundSlice,
        mode: SpectrogramMode = "psd",
    ) -> Spectrogram:
        from maad import sound

        ks = f"{slice_.start}-{slice_.end}"

        if ks in self.__spectrograms[mode].keys():
            return self.__spectrograms[mode][ks]

        s, tn, fn, ext = sound.spectrogram(
            x=np.array(slice_.sound),
            fs=self.sample_rate,
            mode=mode,
        )

        self.__spectrograms[mode][ks] = Spectrogram(s, tn, fn, ext)
        return self.__spectrograms[mode][ks]



from enum import Enum


class MenuChoice(Enum):
    RefreshConfig = "Refresh configuration"
    ExtractAggregate = "Run extractions and aggregations"
    Reduce = "Run reductions"
    ComputeRequirements = (
        "Run computation UMAPs and mean distances matrix"
        " (needed for autocluster and relative trajectories)"
    )
    PurgeRequirements = "Purge computation UMAPs and mean distances matrix"
    Autocluster = "Run autoclusters"
    Trace = "Run trajectories"
    RelativeTrace = "Run relative trajectories"
    Digest = "Run digests"
    RunAll = "Run all"
    ExportDataframe = "Export dataframe as .csv"
    ExportComputationUmaps = "Export computation UMAPs as .npy"
    ExportMeanDistancesMatrix = "Export mean distances matrix as .npy"
    Repack = "Repack storage with `h5repack` (UNIX only)"
    FixAudioWindows10_7_2 = "Fix audio on Windows for versions <=10.7.2"
    Quit = "Quit"



from processing.common.MeanDistancesMatrix import MeanDistancesMatrix
from processing.interfaces import MenuCallback
from processing.storage.Storage import Storage
from processing.utils.invoke_menu import invoke_menu
from processing.utils.print_no_mean_distances_matrices import (
    print_no_mean_distances_matrices,
)


def validate_mean_distances_matrix(action):
    def decorator(
        storage: Storage,
        callback: MenuCallback,
    ):
        if not MeanDistancesMatrix.exists_in_storage(storage):
            print_no_mean_distances_matrices()
            invoke_menu(storage, callback)
            return

        return action(storage, callback)

    return decorator



from typing import List

from processing.config.ConfigParser import ConfigParser
from processing.config.ExcelSheet import ExcelSheet
from processing.config.integrations.IntegrationConfig import IntegrationConfig
from processing.config.integrations.IntegrationExcel import IntegrationExcel
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath
from processing.utils.validate_excel_names import validate_excel_names


class IntegrationStorage:
    names = StoragePath.integrations_names
    seconds = StoragePath.integrations_seconds

    @staticmethod
    def delete_from_storage(storage: Storage) -> None:
        storage.delete(IntegrationStorage.names)
        storage.delete(IntegrationStorage.seconds)

    @staticmethod
    def exists_in_storage(storage: Storage) -> bool:
        return storage.exists_dataset(
            IntegrationStorage.names
        ) and storage.exists_dataset(IntegrationStorage.seconds)

    @staticmethod
    def read_from_storage(storage: Storage) -> List[IntegrationConfig]:
        names_dataset = storage.read(IntegrationStorage.names)

        names = storage.convert_dataset_to_string_list(names_dataset)
        durations = storage.read(IntegrationStorage.seconds)

        integrations = IntegrationConfig.reconstruct(
            names=names,
            durations=durations[:],
        )

        return integrations

    @staticmethod
    def write_to_storage(
        integrations: List[IntegrationConfig],
        storage: Storage,
    ) -> None:
        names, durations = IntegrationConfig.flatten(integrations)
        storage.write(path=IntegrationStorage.names, data=names)
        storage.write(path=IntegrationStorage.seconds, data=durations)

    @staticmethod
    def read_from_config(parser: ConfigParser) -> List[IntegrationConfig]:
        sheet = ExcelSheet.integrations

        names = parser.get(sheet, IntegrationExcel.name_)
        names = validate_excel_names(names)

        durations = parser.get(sheet, IntegrationExcel.duration)

        integrations = IntegrationConfig.reconstruct(
            names=names,
            durations=durations,
        )

        return integrations



import os

import numpy as np

from processing.config.ConfigParser import ConfigParser
from processing.config.ExcelSheet import ExcelSheet
from processing.config.settings.SettingsConfig import SettingsConfig
from processing.config.settings.SettingsDefaults import SettingsDefaults
from processing.config.settings.SettingsRow import SettingsRow
from processing.config.settings.SettingsSheet import SettingsSheet
from processing.constants import STRING_NONE
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath
from processing.utils.is_nan import is_nan


class SettingsStorage:
    settings = StoragePath.settings.value

    @staticmethod
    def delete_from_storage(storage: Storage) -> None:
        storage.delete(SettingsStorage.settings)

    @staticmethod
    def exists_in_storage(storage: Storage) -> bool:
        return storage.exists_dataset(SettingsStorage.settings)

    @staticmethod
    def read_from_storage(storage: Storage) -> SettingsConfig:
        dataset = storage.read(SettingsStorage.settings)
        attributes = dataset.attrs

        storage_path: str = attributes[SettingsRow.storage_path.value]  # type: ignore
        audio_path: str = attributes[SettingsRow.audio_path.value]  # type: ignore
        audio_host: str = attributes[SettingsRow.audio_host.value]  # type: ignore
        expected_sample_rate: int = attributes[
            SettingsRow.expected_sample_rate.value
        ]  # type: ignore
        timeline_origin: int = attributes[
            SettingsRow.timeline_origin.value
        ]  # type: ignore
        timezone: str = attributes[SettingsRow.timezone.value]  # type: ignore
        c_umap_dimensions: int = attributes[
            SettingsRow.c_umap_dimensions.value
        ]  # type: ignore
        c_umap_iterations: int = attributes[
            SettingsRow.c_umap_iterations.value
        ]  # type: ignore
        display_umap_seed: int = attributes[
            SettingsRow.display_umap_seed.value
        ]  # type: ignore

        settings = SettingsConfig(
            storage_path=storage_path,
            audio_path=audio_path,
            audio_host=audio_host,
            expected_sample_rate=expected_sample_rate,
            timeline_origin=timeline_origin,
            timezone=timezone,
            computation_umap_dimensions=c_umap_dimensions,
            computation_umap_iterations=c_umap_iterations,
            display_umap_seed=display_umap_seed,
        )

        return settings

    @staticmethod
    def read_from_config(parser: ConfigParser) -> SettingsConfig:
        sheet = ExcelSheet.settings

        properties = parser.get(sheet, SettingsSheet.setting)
        values = parser.get(sheet, SettingsSheet.value_)

        obj = {}

        for property, value in zip(properties, values):
            obj[property] = value

        storage_path = obj[SettingsRow.storage_path.value]

        if not os.path.isabs(storage_path):
            storage_path = os.path.join(parser.folder, storage_path)

        audio_path = obj[SettingsRow.audio_path.value]

        if not os.path.isabs(audio_path):
            audio_path = os.path.join(parser.folder, audio_path)

        audio_host = obj[SettingsRow.audio_host.value]

        if is_nan(audio_host):
            audio_host = SettingsDefaults.audio_host

        expected_sample_rate = obj[SettingsRow.expected_sample_rate.value]

        timeline_origin = (
            int(obj[SettingsRow.timeline_origin.value].timestamp()) * 1000
        )  # milliseconds

        timezone = (
            obj[SettingsRow.timezone.value]
            if obj[SettingsRow.timezone.value] is not np.nan
            else SettingsDefaults.timezone
        )

        computation_umap_dimensions = (
            obj[SettingsRow.c_umap_dimensions.value]
            if obj[SettingsRow.c_umap_dimensions.value] is not np.nan
            else SettingsDefaults.computation_umap_dimensions
        )

        computation_umap_iterations = (
            obj[SettingsRow.c_umap_iterations.value]
            if obj[SettingsRow.c_umap_iterations.value] is not np.nan
            else SettingsDefaults.computation_umap_iterations
        )

        display_umap_seed = (
            obj[SettingsRow.display_umap_seed.value]
            if obj[SettingsRow.display_umap_seed.value] is not np.nan
            else SettingsDefaults.display_umap_seed
        )

        settings = SettingsConfig(
            storage_path=storage_path,
            audio_path=audio_path,
            audio_host=audio_host,
            expected_sample_rate=expected_sample_rate,
            timeline_origin=timeline_origin,
            timezone=timezone,
            computation_umap_dimensions=computation_umap_dimensions,
            computation_umap_iterations=computation_umap_iterations,
            display_umap_seed=display_umap_seed,
        )

        return settings

    @staticmethod
    def write_to_storage(settings: SettingsConfig, storage: Storage) -> None:
        storage.write_empty_group(SettingsStorage.settings)

        for k, v in vars(settings).items():
            storage.create_attribute(
                path=SettingsStorage.settings,
                key=k,
                value=v if v is not None else STRING_NONE,
            )



from rich.progress import track

from processing.common.TracedStorage import TracedStorage
from processing.config.labels.LabelStorage import LabelStorage
from processing.config.trajectories.TrajectoryStorage import TrajectoryStorage
from processing.interfaces import MenuCallback
from processing.storage.Storage import Storage
from processing.utils.build_aggregated_reduceables import build_aggregated_reduceables
from processing.utils.build_reducers import read_reducers
from processing.utils.invoke_menu import invoke_menu
from processing.utils.print_action import print_action
from processing.utils.print_trajectories import print_trajectories
from processing.utils.validate_aggregated import validate_aggregated
from processing.utils.validate_configuration import validate_configuration
from processing.utils.validate_reduced import validate_reduced


@validate_configuration
@validate_aggregated
@validate_reduced
def trace_trajectories(
    storage: Storage,
    callback: MenuCallback,
):
    print_action("Tracing trajectories started!", "start")

    TracedStorage.delete(storage)

    trajectories = TrajectoryStorage.read_from_storage(storage)
    print_trajectories(trajectories)

    reducers = read_reducers(storage)
    ars = build_aggregated_reduceables(storage)
    labels_properties = LabelStorage.read_properties_from_storage(storage)

    for ar in ars:
        aggregated_timestamps = ar.read_timestamps_from_storage(storage)
        aggregated_labels = ar.read_labels_from_storage(storage)

        for reducer in reducers:
            reducer.load(ar.band, ar.integration)

            if not reducer.should_calculate():
                continue

            reduced_features = ar.read_reduced_from_storage(storage, reducer)

            for trajectory in track(
                trajectories,
                description=(
                    f"Band {ar.band.name}"
                    f", integration {ar.integration.name}"
                    f", reducer {reducer.name}{reducer.dimensions}"
                ),
            ):
                trajectory.create_instance(
                    band=ar.band,
                    integration=ar.integration,
                    reducer=reducer,
                )

                trajectory.instance.load(
                    features=reduced_features,
                    timestamps=aggregated_timestamps,
                    timestamp_start=trajectory.start,
                    timestamp_end=trajectory.end,
                    labels_properties=labels_properties,
                    labels_values=aggregated_labels,
                    step=trajectory.step,
                )

                trajectory.instance.calculate(
                    trajectory_label_property=trajectory.label_property,
                    trajectory_label_value=trajectory.label_value,
                )

                TracedStorage.write_data(storage, trajectory, reducer, ar)
                TracedStorage.write_timestamps(storage, trajectory, reducer, ar)
                TracedStorage.write_relative_timestamps(
                    storage,
                    trajectory,
                    reducer,
                    ar,
                )

    print_action("Tracing trajectories completed!", "end")
    invoke_menu(storage, callback)



from processing.extractors.Extractor import Extractor
from processing.loaders.Loader import Loader


class FrequencyEntropyExtractor(Extractor):
    """hf"""

    def extract(self, loader: Loader):
        from maad import features

        data = []

        for slice_ in self.sound_walk(loader):
            spectrogram = loader.sound.get_spectrogram(slice_)
            hf, _ = features.frequency_entropy(spectrogram.s)
            data.append([hf])

        return data



from rich.console import Console
from rich.table import Table

from processing.utils.pack_trajectories import PackedTrajectories
from processing.utils.walk_packed_trajectories import walk_packed_trajectories


def print_packed_trajectories(packed_trajectories: PackedTrajectories):
    console = Console()

    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("Pack number")
    table.add_column("Label property")
    table.add_column("Label value")
    table.add_column("Trajectories count")

    index = 1

    for label_property, label_value, pack in walk_packed_trajectories(
        packed_trajectories
    ):
        table.add_row(
            str(index),
            str(label_property),
            str(label_value),
            str(len(packed_trajectories[label_property][label_value])),
        )

        index += 1

    console.print(table)



from typing import List, Tuple

from processing.config.files.FileSheet import FileSheet
from processing.utils.get_uniques_sorted import get_uniques_sorted
from processing.utils.get_uniques_unsorted import get_uniques_unsorted
from processing.utils.reverse_array import reverse_array


class LabelConfig:
    index: int
    property: str
    values: List[str]  # label values by file index
    uniques_sorted: List[str]  # by alphanumerical order
    uniques_unsorted: List[str]  # by order of occurrence

    def __init__(
        self,
        index: int,
        property: str,
    ) -> None:
        self.index = index
        self.property = property

    @staticmethod
    def is_label_property(string: str) -> bool:
        prefix = FileSheet.label_prefix.value
        prefix_length = len(prefix)
        string_slice = string[:prefix_length]

        if prefix == string_slice:
            return True

        return False

    @staticmethod
    def convert_to_values_by_file(
        metas: List["LabelConfig"],
    ) -> List[List[str]]:
        meta_values = [m.values for m in metas]
        return reverse_array(meta_values)

    @staticmethod
    def flatten(
        metas: List["LabelConfig"],
    ) -> Tuple[List[str], List[List[str]]]:
        properties = [str.upper(m.property) for m in metas]
        sets = [list(m.uniques_sorted) for m in metas]

        return properties, sets

    @staticmethod
    def trim_prefixed_property_from_config(string: str) -> str:
        assert LabelConfig.is_label_property(
            string
        ), f"Unable to find label prefix in string {string}."

        return string.replace(FileSheet.label_prefix.value, "")

    def load_values(
        self,
        array: List,
    ) -> List[str]:
        self.values = [str(value) for value in array]
        self.uniques_sorted = get_uniques_sorted(self.values)
        self.uniques_unsorted = get_uniques_unsorted(self.values)
        return self.values

    def get_value(
        self,
        file_index: int,
    ) -> str:
        return self.values[file_index]



from processing.config.Config import Config
from processing.interfaces import MenuCallback
from processing.storage.Storage import Storage
from processing.utils.invoke_menu import invoke_menu
from processing.utils.print_action import print_action


def refresh_configuration(
    config: Config,
    storage: Storage,
    callback: MenuCallback = None,
):
    print_action("Configuration refresh started!", "start")

    config = Config(path=config.path)
    config.write(storage)

    print_action("Configuration refresh completed!", "end")
    invoke_menu(storage, callback)



from typing import List

from sklearn.decomposition import PCA
from sklearn.preprocessing import scale

from processing.reducers.AbstractReducer import AbstractReducer


class PcaReducer(AbstractReducer):
    def calculate(self) -> List[List[float]]:
        features = self._validate_load()

        scaled_features = scale(features)

        pca = PCA(
            n_components=self._dimensions,
            random_state=self._seed,
        )

        reduced_features = pca.fit_transform(scaled_features)

        self.set_values(list(reduced_features))

        return self.values



from rich.progress import track

from processing.common.AggregatedReduceable import AggregatedReduceable
from processing.config.bands.BandStorage import BandStorage
from processing.config.extractors.ExtractorStorage import ExtractorStorage
from processing.config.integrations.IntegrationStorage import IntegrationStorage
from processing.config.ranges.RangeStorage import RangeStorage
from processing.config.reducers.ReducerStorage import ReducerStorage
from processing.config.settings.SettingsStorage import SettingsStorage
from processing.interfaces import MenuCallback
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath
from processing.utils.filter_nn_extractors import filter_nn_extractors
from processing.utils.invoke_menu import invoke_menu
from processing.utils.print_action import print_action
from processing.utils.print_aggregated_reduceables import print_aggregated_reduceables
from processing.utils.print_no_aggregated_reduceables import (
    print_no_aggregated_reduceables,
)
from processing.utils.print_reducers import print_reducers
from processing.utils.validate_aggregated import validate_aggregated
from processing.utils.validate_configuration import validate_configuration


@validate_configuration
@validate_aggregated
def reduce(
    storage: Storage,
    callback: MenuCallback,
):
    print_action("Reductions started!", "start")

    storage.delete(StoragePath.reduced)

    settings = SettingsStorage.read_from_storage(storage)
    bands = BandStorage.read_from_storage(storage)
    integrations = IntegrationStorage.read_from_storage(storage)
    ranges = RangeStorage.read_from_storage(storage)
    reducers = ReducerStorage.read_from_storage(storage, bands, integrations, ranges)
    extractors = ExtractorStorage.read_from_storage(storage)
    nn_extractors = filter_nn_extractors(extractors)

    aggregated_reduceables = AggregatedReduceable.reconstruct(
        bands=bands,
        integrations=integrations,
        nn_extractors=nn_extractors,
    )

    if len(aggregated_reduceables) == 0:
        print_no_aggregated_reduceables()
        if callback is not None:
            callback(storage)
        return

    print_reducers(reducers)
    print_aggregated_reduceables(aggregated_reduceables, storage)

    for ar in track(aggregated_reduceables, description="Reducing..."):
        if not ar.exists_in_storage(storage):
            continue

        features = ar.read_features_from_storage(storage)

        for reducer in reducers:
            reducer.create_instance(ar.band, ar.integration)

            if not reducer.should_calculate():
                continue

            reducer.instance.load(
                dimensions=reducer.dimensions,
                seed=settings.display_umap_seed,
                features=features[:],
            )

            reducer.instance.calculate()

            storage.append(
                path=ar.get_reduced_path(reducer),
                data=reducer.instance.values,
                compression=True,
                attributes={
                    "extractor": ar.extractor.name,
                    "extractor_index": str(ar.extractor.index),
                    "reducer": reducer.name,
                    "reducer_index": str(reducer.index),
                    "reducer_dimensions": str(reducer.dimensions),
                },
            )

    print_action("Reductions completed!", "end")
    invoke_menu(storage, callback)



from enum import Enum


class ExcelSheet(Enum):
    help = "Help"
    help_extractors = "HelpExtractors"
    help_reducers = "HelpReducers"
    help_autoclusters = "HelpAutoclusters"
    help_digesters = "HelpDigesters"
    settings = "Settings"
    files = "Files"
    bands = "Bands"
    integrations = "Integrations"
    ranges = "Ranges"
    extractors = "Extractors"
    reducers = "Reducers"
    autoclusters = "Autoclusters"
    digesters = "Digesters"
    trajectories = "Trajectories"



from processing.extractors.Extractor import Extractor
from processing.loaders.Loader import Loader


class TemporalEntropyExtractor(Extractor):
    """ht"""

    def extract(self, loader: Loader):
        from maad import features

        data = []

        for slice_ in self.sound_walk(loader):
            temporal_entropy = features.temporal_entropy(slice_.sound)
            data.append([temporal_entropy])

        return data



from typing import Dict, List

from processing.common.FileLoader import FileLoader
from processing.common.FileWalker import FileWalker
from processing.common.Interval import Block, Interval
from processing.config.files.FileConfig import FileConfig
from processing.config.integrations.IntegrationConfig import IntegrationConfig
from processing.config.sites.SiteConfig import SiteConfig
from processing.errors.TimelineIntervalOverlapError import TimelineIntervalOverlapError
from processing.storage.Storage import Storage

TimelineMap = Dict[int, Interval]
FileIndex = int


# One timeline per site
class Timeline:
    def __init__(
        self,
        site: SiteConfig,
        integration: IntegrationConfig,
        storage: Storage,
        origin: int,
        debug: bool = False,
    ) -> None:
        self.site = site
        self.integration = integration
        self.storage = storage
        self.origin = origin
        self.debug = debug

        self.map: TimelineMap = {}
        self.step = integration.milliseconds
        self.loaded_files: List[FileConfig] = []
        self.efs: Dict[FileIndex, FileLoader] = {}

        self.load()
        self.print_debug()

    def get_interval_count_in_file(self, file: FileConfig) -> int:
        count = (file.end - file.start) // self.step

        if file.duration < self.step:
            count += 1

        before_remainder = (file.start - self.origin) % self.step

        before = file.start - before_remainder
        after = before + count * self.step

        if file.end > after:
            count += 1

        return count

    def verify_overlaps(
        self,
        interval: Interval,
        block: Block,
    ) -> None:
        """Return true if overlap"""
        existing_interval = self.map[interval.start]
        for existing_block in existing_interval.blocks:
            if block.start <= existing_block.end or block.end <= existing_block.start:
                raise TimelineIntervalOverlapError(block.file, existing_block.file)

    def add_interval(self, interval: Interval) -> TimelineMap:
        self.map[interval.start] = interval
        self.map = dict(sorted(self.map.items()))
        return self.map

    def load(self):
        for file in self.site.files:
            fw = FileWalker(
                file=file,
                origin=self.origin,
                step=self.step,
                debug=self.debug,
            )

            for interval, block in fw.walk():
                # create in map if not present
                if interval.start not in self.map:
                    self.add_interval(interval)

                # verify overlaps if present
                if interval.start in self.map:
                    self.verify_overlaps(interval, block)

                self.map[interval.start].blocks.append(block)

    def print_debug(self):
        if self.debug is False:
            return

        for interval_start, interval in self.map.items():
            print(
                interval_start,
                [f"{b.file.index}-{b.start}-{b.end}" for b in interval.blocks],
            )

    @property
    def intervals(self):
        return self.map.values()



from typing import List

from processing.config.bands.BandStorage import BandStorage
from processing.config.integrations.IntegrationStorage import IntegrationStorage
from processing.config.ranges.RangeStorage import RangeStorage
from processing.config.reducers.ReducerConfig import ReducerConfig
from processing.config.reducers.ReducerStorage import ReducerStorage
from processing.storage.Storage import Storage


def read_reducers(
    storage: Storage,
) -> List[ReducerConfig]:
    bands = BandStorage.read_from_storage(storage)
    integrations = IntegrationStorage.read_from_storage(storage)
    ranges = RangeStorage.read_from_storage(storage)
    reducers = ReducerStorage.read_from_storage(storage, bands, integrations, ranges)
    return reducers



import numpy as np
from pandas import DataFrame

from processing.config.labels import LabelConfig


def sort_dataframe(dataframe: DataFrame, label: LabelConfig) -> DataFrame:
    order = np.argsort(label.uniques_unsorted)
    dataframe = dataframe[:, order]
    dataframe = dataframe[order, :]
    return dataframe



from rich import print


def print_menu_legend():
    print("State in storage:")
    print("[red]❌ Missing[/red]")
    print("[green]✅ Present[/green]")
    print()

    print(
        "[yellow]"
        "⚠️ Remember to [b]Quit[/b] before loading file into the visualisation module"
        "[/yellow]"
    )
    print()



from typing import List

from sklearn.decomposition import SparsePCA

from processing.reducers.AbstractReducer import AbstractReducer


class SparsePcaReducer(AbstractReducer):
    def calculate(
        self,
    ) -> List[List[float]]:
        features = self._validate_load()

        sparse_pca = SparsePCA(
            n_components=self._dimensions,
            random_state=self._seed,
        )

        reduced_features = sparse_pca.fit_transform(features)
        self.set_values(list(reduced_features))
        return self.values



from typing import List, Literal

import numpy as np
import torch
from torch import Tensor, cuda, hub, nn

from processing.extractors.Extractor import Extractor
from processing.loaders.Loader import Loader


class VggExtractor(nn.Module, Extractor):
    def __init__(self) -> None:
        super(VggExtractor, self).__init__()

        self.features: nn.Sequential
        self.embeddings: nn.Sequential
        self.device: Literal["cuda", "cpu"]

        # Hardcoding vgg behaviour
        # This is a reflection of the current implementation.
        # `offset` and `step` are not actually read.
        # This is for abstract parent class validation
        self.mel_count: int = 64
        self.fft_size: int = 2048
        self.offset = 0
        self.step = 1000

        self.load()

    # Loading model
    def load(self):
        self.features = self.load_features()
        self.embeddings = self.load_embeddings()
        self.device = self.load_device()
        self.load_state_dict()
        self.to(self.device)
        self.eval()

    def load_embeddings(self):
        return nn.Sequential(
            nn.Linear(512 * 4 * 6, 4096),
            nn.ReLU(True),
            nn.Linear(4096, 4096),
            nn.ReLU(True),
            nn.Linear(4096, 128),
            nn.ReLU(True),
        )

    def load_features(self):
        layers = []
        in_channels = 1

        for v in [64, "M", 128, "M", 256, 256, "M", 512, 512, "M"]:
            if v == "M":
                layers += [nn.MaxPool2d(kernel_size=2, stride=2)]
            else:
                conv2d = nn.Conv2d(in_channels, v, kernel_size=3, padding=1)
                layers += [conv2d, nn.ReLU(inplace=True)]
                in_channels = v

        return nn.Sequential(*layers)

    def load_state_dict(self):
        state_dict = self.fetch_state_dict()
        super().load_state_dict(state_dict)

    @staticmethod
    def fetch_state_dict():
        return hub.load_state_dict_from_url(
            "https://github.com/harritaylor/torchvggish/"
            "releases/download/v0.1/vggish-10086976.pth",
            progress=True,
        )

    def load_device(self):
        if cuda.is_available():
            print("Using GPU")
            return "cuda"
        else:
            print("Using CPU")
            return "cpu"

    def forward(
        self,
        x: Tensor,
    ) -> Tensor:
        with torch.no_grad():
            x = self.features(x)
            # Transpose the output from features to
            # remain compatible with VGGish embeddings
            x = torch.transpose(x, 1, 3)
            x = torch.transpose(x, 1, 2)
            x = x.contiguous()
            x = x.view(x.size(0), -1)
            return self.embeddings(x)

    def to_frames(self, data: Tensor) -> Tensor:
        window_length = hop_length = 100

        num_samples = data.shape[0]
        num_frames = 1 + int(np.floor((num_samples - window_length) / hop_length))
        shape = (num_frames, window_length) + data.shape[1:]

        strides = data.stride(0) * hop_length, data.stride(0), data.stride(1)

        return torch.as_strided(data, shape, strides)

    def convert_to_mel(
        self,
        wav: Tensor,
        sample_rate: int,
        hop_length: int,
    ) -> Tensor:
        import torchaudio

        data = wav.reshape(-1)

        mel_extractor = torchaudio.transforms.MelSpectrogram(
            sample_rate=sample_rate,
            n_fft=self.fft_size,
            hop_length=hop_length,
            f_min=self.band.low,
            f_max=self.band.high,
            n_mels=self.mel_count,
            win_length=self.fft_size,
            power=1,
        )

        mel_extractor = mel_extractor.to(data.device)

        mel = mel_extractor(data)
        log: Tensor = torch.log(mel + 0.1)
        transposed_log: Tensor = log.T

        # Frame features into examples.
        mel_frames = self.to_frames(data=transposed_log)

        mel_frames = mel_frames[:, None, :, :].float()
        return mel_frames

    def extract(self, loader: Loader):
        wav = loader.torch.audio
        sample_rate = loader.torch.sample_rate

        assert wav is not None and sample_rate is not None, "not loaded"
        assert sample_rate == self.expected_sample_rate, "Sample rates do not match."

        # `batch` acts as the maximum audio duration of 5 minutes.
        batch = sample_rate * 60 * 5
        tensors: List[Tensor] = []
        hop_length = sample_rate // 100

        i = 0
        while i < wav.shape[1]:
            start = i
            end = i + batch
            wav_slice: Tensor = wav[:, start:end]
            wav_slice_converted = wav_slice.to(self.device)
            mel_slice = self.convert_to_mel(
                wav_slice_converted,
                sample_rate,
                hop_length,
            )
            tensor = self.forward(mel_slice)
            tensors.append(tensor)
            i += batch

        features = torch.concat(tensors)
        return features.tolist()



from typing import List

from processing.config.bands.BandConfig import BandConfig
from processing.config.ConfigParser import ConfigParser
from processing.config.ExcelSheet import ExcelSheet
from processing.config.integrations.IntegrationConfig import IntegrationConfig
from processing.config.ranges.RangeConfig import RangeConfig
from processing.config.reducers.ReducerConfig import ReducerConfig
from processing.config.reducers.ReducerExcel import ReducerExcel
from processing.constants import STRING_NONE
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath
from processing.utils.is_nan import is_nan


class ReducerStorage:
    names = StoragePath.reducers_names.value
    dimensions = StoragePath.reducers_dimensions.value
    ranges = StoragePath.reducers_ranges.value
    bands = StoragePath.reducers_bands.value
    integrations = StoragePath.reducers_integrations.value

    @staticmethod
    def delete_from_storage(storage: Storage) -> None:
        storage.delete(ReducerStorage.names)
        storage.delete(ReducerStorage.dimensions)
        storage.delete(ReducerStorage.ranges)
        storage.delete(ReducerStorage.bands)
        storage.delete(ReducerStorage.integrations)

    @staticmethod
    def exists_in_storage(storage: Storage) -> bool:
        return (
            storage.exists_dataset(ReducerStorage.names)
            and storage.exists_dataset(ReducerStorage.dimensions)
            and storage.exists_dataset(ReducerStorage.ranges)
            and storage.exists_dataset(ReducerStorage.bands)
            and storage.exists_dataset(ReducerStorage.integrations)
        )

    @staticmethod
    def read_from_storage(
        storage: Storage,
        bands: List[BandConfig],
        integrations: List[IntegrationConfig],
        ranges: List[RangeConfig],
    ) -> List[ReducerConfig]:
        names_dataset = storage.read(ReducerStorage.names)

        names = storage.convert_dataset_to_string_list(names_dataset)
        dimensions = storage.read(ReducerStorage.dimensions)[:]

        bands_names = storage.read(ReducerStorage.bands).asstr()
        integrations_names = storage.read(ReducerStorage.integrations).asstr()
        ranges_names = storage.read(ReducerStorage.ranges).asstr()

        bands_names = storage.trim_rectangular(bands_names, STRING_NONE)
        integrations_names = storage.trim_rectangular(integrations_names, STRING_NONE)
        ranges_names = storage.trim_rectangular(ranges_names, STRING_NONE)

        bands_names_strings = [
            ",".join(bands_names_string) for bands_names_string in bands_names
        ]

        integrations_names_strings = [
            ",".join(integrations_names_string)
            for integrations_names_string in integrations_names
        ]

        ranges_names_strings = [
            ",".join(ranges_names_string) for ranges_names_string in integrations_names
        ]

        reducers = ReducerConfig.reconstruct(
            names=names,
            dimensions=dimensions,
            bands_names_strings=bands_names_strings,
            integrations_names_strings=integrations_names_strings,
            ranges_names_strings=ranges_names_strings,
            bands=bands,
            integrations=integrations,
            ranges=ranges,
        )

        return reducers

    @staticmethod
    def read_from_config(
        parser: ConfigParser,
        bands: List[BandConfig],
        integrations: List[IntegrationConfig],
        ranges: List[RangeConfig],
    ) -> List[ReducerConfig]:
        sheet = ExcelSheet.reducers

        names = parser.get(sheet, ReducerExcel.name_)
        dimensions = parser.get(sheet, ReducerExcel.dimensions)

        for d in dimensions:
            assert is_nan(d) is False, f"{d} is not a valid dimension"

        bands_names_string = parser.get(sheet, ReducerExcel.bands)

        integrations_names_string = parser.get(
            sheet,
            ReducerExcel.integrations,
        )

        ranges_names_strings = parser.get(sheet, ReducerExcel.ranges)

        reducers = ReducerConfig.reconstruct(
            names=names,
            dimensions=dimensions,
            bands_names_strings=bands_names_string,
            integrations_names_strings=integrations_names_string,
            ranges_names_strings=ranges_names_strings,
            bands=bands,
            integrations=integrations,
            ranges=ranges,
        )

        return reducers

    @staticmethod
    def write_to_storage(reducers: List[ReducerConfig], storage: Storage) -> None:
        (
            names,
            dimensions,
            bands,
            integrations,
            ranges,
        ) = ReducerConfig.flatten(reducers=reducers)

        if len(bands) != 0:
            bands = storage.make_rectangular(bands, "")

        if len(integrations) != 0:
            integrations = storage.make_rectangular(integrations, "")

        if len(ranges) != 0:
            ranges = storage.make_rectangular(ranges, "")

        storage.write(path=ReducerStorage.names, data=names)
        storage.write(path=ReducerStorage.dimensions, data=dimensions)
        storage.write(path=ReducerStorage.bands, data=bands)
        storage.write(path=ReducerStorage.integrations, data=integrations)
        storage.write(path=ReducerStorage.ranges, data=ranges)



from typing import List

from h5py import Dataset

from processing.config.bands.BandConfig import BandConfig
from processing.config.extractors.ExtractorConfig import ExtractorConfig
from processing.config.integrations.IntegrationConfig import IntegrationConfig
from processing.config.reducers.ReducerConfig import ReducerConfig
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath
from processing.utils.walk_bands_integrations import walk_bands_integrations


class AggregatedReduceable:
    def __init__(
        self,
        extractor: ExtractorConfig,
        band: BandConfig,
        integration: IntegrationConfig,
    ) -> None:
        self.extractor = extractor
        self.band = band
        self.integration = integration

        self.path = (
            f"{StoragePath.aggregated.value}"
            f"/{self.band.name}"
            f"/{self.integration.seconds}"
            f"/{self.extractor.index}"
        )

    def read_features_from_storage(self, storage: Storage) -> Dataset:
        return storage.read(self.path)

    def read_timestamps_from_storage(self, storage: Storage) -> Dataset:
        path = self.get_timestamps_path()
        return storage.read(path)

    def read_labels_from_storage(self, storage: Storage) -> Dataset:
        path = self.get_labels_path()
        return storage.read(path)

    def read_reduced_from_storage(
        self,
        storage: Storage,
        reducer: ReducerConfig,
    ) -> Dataset:
        path = self.get_reduced_path(reducer)
        return storage.read(path)

    def exists_in_storage(self, storage: Storage) -> bool:
        return storage.exists_dataset(self.path)

    def get_timestamps_path(self) -> str:
        return (
            f"{StoragePath.aggregated_timestamps.value}"
            f"/{self.band.name}"
            f"/{self.integration.seconds}"
            f"/{self.extractor.index}"
        )

    def get_labels_path(self) -> str:
        return (
            f"{StoragePath.aggregated_labels.value}"
            f"/{self.band.name}"
            f"/{self.integration.seconds}"
            f"/{self.extractor.index}"
        )

    def get_reduced_path(self, reducer: ReducerConfig) -> str:
        return (
            f"{StoragePath.reduced.value}"
            f"/{self.band.name}"
            f"/{self.integration.seconds}"
            f"/{self.extractor.index}"
            f"/{reducer.index}"
        )

    @staticmethod
    def reconstruct(
        bands: List[BandConfig],
        integrations: List[IntegrationConfig],
        nn_extractors: List[ExtractorConfig],
    ) -> List["AggregatedReduceable"]:
        aggregated_reduceables: List[AggregatedReduceable] = []

        for nn_ex in nn_extractors:
            for band, integration in walk_bands_integrations(bands, integrations):
                aggregated_reduceable = AggregatedReduceable(
                    extractor=nn_ex,
                    band=band,
                    integration=integration,
                )

                aggregated_reduceables.append(aggregated_reduceable)

        return aggregated_reduceables



import numpy
from rich.progress import track
from sklearn.neighbors import NearestNeighbors

from processing.common.ComputationUmapStorage import ComputationUmapStorage
from processing.common.RelativeTracedStorage import RelativeTracedStorage
from processing.config.labels.LabelStorage import LabelStorage
from processing.config.trajectories.TrajectoryStorage import TrajectoryStorage
from processing.interfaces import MenuCallback
from processing.storage.Storage import Storage
from processing.utils.build_aggregated_reduceables import build_aggregated_reduceables
from processing.utils.compute_relative_distances import compute_relative_distances
from processing.utils.compute_starting_point import compute_starting_point
from processing.utils.invoke_menu import invoke_menu
from processing.utils.pack_trajectories import pack_trajectories
from processing.utils.print_action import print_action
from processing.utils.print_aggregated_reduceables import print_aggregated_reduceables
from processing.utils.print_packed_trajectories import print_packed_trajectories
from processing.utils.read_trajectory_path_and_relative_timestamps import (
    read_trajectory_path_and_relative_timestamps,
)
from processing.utils.validate_computation_umap import validate_computation_umap
from processing.utils.validate_configuration import validate_configuration
from processing.utils.walk_packed_trajectories import walk_packed_trajectories


@validate_configuration
@validate_computation_umap
def trace_relative_trajectories(
    storage: Storage,
    callback: MenuCallback,
) -> None:
    print_action("Tracing relative trajectories started!", "start")

    RelativeTracedStorage.delete(storage)

    trajectories = TrajectoryStorage.read_from_storage(storage)
    packs = pack_trajectories(trajectories)
    print_packed_trajectories(packs)

    ars = build_aggregated_reduceables(storage)
    print_aggregated_reduceables(ars, storage)

    labels_properties = LabelStorage.read_properties_from_storage(storage)
    knner = NearestNeighbors(n_neighbors=100)

    for ar in ars:
        ar_timestamps = ar.read_timestamps_from_storage(storage)
        ar_labels = ar.read_labels_from_storage(storage)

        computation_umaps = ComputationUmapStorage.read_from_storage(
            storage=storage,
            band=ar.band,
            integration=ar.integration,
        )

        # a pack is a list of trajectories linked by the same label
        for label_property, label_value, trajectories in walk_packed_trajectories(
            packs
        ):
            pack_relative_distances = []
            pack_relative_timestamps = []

            # building arrays to populate
            for _ in enumerate(trajectories):
                pack_relative_distances.append([])
                pack_relative_timestamps.append([])

            # iterating through computation UMAPs
            for computation_umap in track(
                computation_umaps,
                description=f"Tracing {label_property}: {label_value}",
            ):
                paths = []

                # reading path for each trajectory because all are needed to compute
                # the starting point
                for t, trajectory in enumerate(trajectories):
                    (
                        path,
                        relative_timestamps,
                    ) = read_trajectory_path_and_relative_timestamps(
                        trajectory=trajectory,
                        features=computation_umap,
                        timestamps=ar_timestamps,
                        labels_properties=labels_properties,
                        labels_values=ar_labels,
                        label_property=label_property,
                        label_value=label_value,
                    )

                    paths.append(path)
                    pack_relative_timestamps[t].append(relative_timestamps)

                starting_point = compute_starting_point(paths)

                knner.fit(computation_umap)
                dknn, _ = knner.kneighbors(starting_point)
                mean_distance = numpy.mean(dknn)

                for t, _ in enumerate(trajectories):
                    relative_distances = compute_relative_distances(
                        path=paths[t],
                        starting_point=starting_point,
                        mean_distance=mean_distance,
                    )

                    pack_relative_distances[t].append(relative_distances)

            # median arrays and write to storage
            for t, trajectory in enumerate(trajectories):
                to_store = numpy.median(pack_relative_distances[t], axis=0)
                to_store2 = numpy.median(pack_relative_timestamps[t], axis=0)

                RelativeTracedStorage.write_distances(
                    storage=storage,
                    trajectory=trajectory,
                    ar=ar,
                    relative_distances=to_store,
                    label_property=label_property,
                    label_value=label_value,
                )

                RelativeTracedStorage.write_timestamps(
                    storage=storage,
                    trajectory=trajectory,
                    ar=ar,
                    relative_timestamps=to_store2,
                )

    print_action("Tracing relative trajectories completed!", "end")
    invoke_menu(storage, callback)



from typing import List

from processing.config.files.FileConfig import FileConfig
from processing.config.files.FileStorage import FileStorage
from processing.config.settings.SettingsConfig import SettingsConfig
from processing.config.sites.SiteConfig import SiteConfig
from processing.constants import INT_NONE
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath


class SiteStorage:
    path_names = StoragePath.sites_names.value
    path_file_indexes = StoragePath.sites_file_indexes.value

    @staticmethod
    def delete_from_storage(storage: Storage) -> None:
        storage.delete(SiteStorage.path_names)
        storage.delete(SiteStorage.path_file_indexes)

    @staticmethod
    def exists_in_storage(storage: Storage) -> bool:
        return storage.exists_dataset(
            SiteStorage.path_names
        ) and storage.exists_dataset(SiteStorage.path_file_indexes)

    @staticmethod
    def read_from_storage(
        storage: Storage,
        settings: SettingsConfig,
    ) -> List[SiteConfig]:
        names_dataset = storage.read(SiteStorage.path_names)
        names: List[str] = storage.convert_dataset_to_string_list(names_dataset)

        file_indexes_rectangular = storage.read(SiteStorage.path_file_indexes)[:]

        file_indexes: List[List[int]] = storage.trim_rectangular(
            file_indexes_rectangular,
            INT_NONE,
        )

        files = FileStorage.read_from_storage(storage, settings)

        sites = SiteConfig.reconstruct(
            names=names,
            file_indexes=file_indexes,
            files=files,
        )

        return sites

    @staticmethod
    def write_to_storage(sites: List[SiteConfig], storage: Storage) -> None:
        names, file_indexes = SiteConfig.flatten(sites)

        storage.write(
            path=SiteStorage.path_names,
            data=names,
        )

        file_indexes_rectangular = storage.make_rectangular(file_indexes, INT_NONE)

        storage.write(
            path=SiteStorage.path_file_indexes,
            data=file_indexes_rectangular,
        )

    @staticmethod
    def parse_from_config(files: List[FileConfig]) -> List[SiteConfig]:
        # Listing unique site names
        # Making this by hand because using `set()` has inconsistent order
        site_names = []
        for file in files:
            if file.site in site_names:
                continue

            site_names.append(file.site)

        sites = []

        for site_index, site_name in enumerate(site_names):
            picked_files = [file for file in files if file.site == site_name]

            site = SiteConfig(
                index=site_index,
                name=site_name,
                files=picked_files,
            )

            sites.append(site)

        return sites



import os
from typing import List

from pandas import DataFrame, ExcelWriter
from rich import print

from processing.config.ConfigParser import ConfigParser
from processing.config.ExcelSheet import ExcelSheet
from processing.config.files.FileSheet import FileSheet


def append_to_config(config_path: str, paths: List[str]) -> None:
    folder_path = os.path.dirname(config_path)
    parser = ConfigParser(config_path, folder_path)
    filenames = parser.get(ExcelSheet.files, FileSheet.name_)

    files_to_append = [path for path in paths if path not in filenames]

    if len(files_to_append) == 0:
        print("No files to append")
        return

    with ExcelWriter(
        path=config_path,
        if_sheet_exists="overlay",
        mode="a",
    ) as writer:
        df = DataFrame([*filenames, *files_to_append], columns=[FileSheet.name_.value])
        df.to_excel(
            writer,
            sheet_name=ExcelSheet.files.value,
            index=False,
        )
        print(f"Files appended to config: {len(files_to_append)}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb  1 18:22:30 2023

@author: rj102145
"""

import numpy as np
import pandas as pd
import torch
import torch.nn as nn
import torch.nn.functional as F
from sklearn import preprocessing
from torch import optim
from torch.autograd import Variable
from torch.utils.data import DataLoader, Dataset

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


def load_data(x):
    df = pd.DataFrame(x)
    # replace nan with -99
    df = df.fillna(-99)
    x = df.values.reshape(-1, df.shape[1]).astype("float32")
    # stadardize values
    standardizer = preprocessing.StandardScaler()
    x = standardizer.fit_transform(x)
    return x, standardizer


def numpy_to_tensor(x):
    x_train = torch.from_numpy(x).to(device)
    return x_train


class DataBuilder(Dataset):
    def __init__(self, x):
        self.x, self.standardizer = load_data(x)
        self.x = numpy_to_tensor(self.x)
        self.len = self.x.shape[0]

    def __getitem__(self, index):
        return self.x[index]

    def __len__(self):
        return self.len


class Autoencoder(nn.Module):
    def __init__(
        self,
        d_in,
        h=[64, 32],  # INFO: This argument is mutable.
        latent_dim=2,
    ):
        # Encoder
        super(Autoencoder, self).__init__()
        # input layer
        self.linear1 = nn.Linear(d_in, h[0])
        self.lin_bn1 = nn.BatchNorm1d(num_features=h[0])

        # hidden layers
        self.HiddenLayersIn = []
        for i in range(len(h) - 1):
            self.linear2 = nn.Linear(h[i], h[i + 1]).to(device)
            self.lin_bn2 = nn.BatchNorm1d(num_features=h[i + 1]).to(device)
            self.HiddenLayersIn.append([self.linear2, self.lin_bn2])

        self.linear3 = nn.Linear(h[-1], h[-1])
        self.lin_bn3 = nn.BatchNorm1d(num_features=h[-1])

        #         # Latent vectors mu and sigma
        self.fc1 = nn.Linear(h[-1], latent_dim)
        self.bn1 = nn.BatchNorm1d(num_features=latent_dim)
        self.fc21 = nn.Linear(latent_dim, latent_dim)
        self.fc22 = nn.Linear(latent_dim, latent_dim)

        #         # Sampling vector
        self.fc3 = nn.Linear(latent_dim, latent_dim)
        self.fc_bn3 = nn.BatchNorm1d(latent_dim)
        self.fc4 = nn.Linear(latent_dim, h[-1])
        self.fc_bn4 = nn.BatchNorm1d(h[-1])

        #         # Decoder
        self.linear4 = nn.Linear(h[-1], h[-1])
        self.lin_bn4 = nn.BatchNorm1d(num_features=h[-1])

        self.HiddenLayersOut = []
        for i in range(len(h) - 1, 0, -1):
            self.linear5 = nn.Linear(h[i], h[i - 1]).to(device)
            self.lin_bn5 = nn.BatchNorm1d(num_features=h[i - 1]).to(device)
            self.HiddenLayersOut.append([self.linear5, self.lin_bn5])

        # output layer
        self.linear6 = nn.Linear(h[0], d_in)
        self.lin_bn6 = nn.BatchNorm1d(num_features=d_in)

        self.relu = nn.ReLU()

    def encode(self, x):
        lin1 = self.relu(self.lin_bn1(self.linear1(x)))
        lin_last = lin1
        for i in range(len(self.HiddenLayersIn)):
            lin_last = self.relu(
                self.HiddenLayersIn[i][1](self.HiddenLayersIn[i][0](lin_last))
            )
        lin3 = self.relu(self.lin_bn3(self.linear3(lin_last)))

        fc1 = F.relu(self.bn1(self.fc1(lin3)))
        r1 = self.fc21(fc1)
        r2 = self.fc22(fc1)

        return r1, r2

    def reparameterize(self, mu, logvar):
        if self.training:
            std = logvar.mul(0.5).exp_()
            eps = Variable(std.data.new(std.size()).normal_())
            return eps.mul(std).add_(mu)
        else:
            return mu

    def decode(self, z):
        fc3 = self.relu(self.fc_bn3(self.fc3(z)))

        fc4 = self.relu(self.fc_bn4(self.fc4(fc3)))
        lin4 = self.relu(self.lin_bn4(self.linear4(fc4)))
        lin_last = lin4
        for i in range(len(self.HiddenLayersOut)):
            lin_last = self.relu(
                self.HiddenLayersOut[i][1](self.HiddenLayersOut[i][0](lin_last))
            )
        return self.lin_bn6(self.linear6(lin_last))

    def forward(self, x):
        mu, logvar = self.encode(x)
        z = self.reparameterize(mu, logvar)
        return self.decode(z), mu, logvar


class custom_loss(nn.Module):
    def __init__(self):
        super(custom_loss, self).__init__()
        self.mse_loss = nn.MSELoss(reduction="sum")

    def forward(self, x_recon, x, mu, logvar):
        loss_mse = self.mse_loss(x_recon, x)
        loss_kld = -0.5 * torch.sum(1 + logvar - mu.pow(2) - logvar.exp())

        return loss_mse + loss_kld


def weights_init_uniform_rule(m):
    # TODO : but why a uniform init? why not a random one with a fixed seed
    classname = m.__class__.__name__
    # for every Linear layer in a model..
    if classname.find("Linear") != -1:
        # get the number of the inputs
        n = m.in_features
        y = 1.0 / np.sqrt(n)
        m.weight.data.uniform_(-y, y)
        m.bias.data.fill_(0)


def train(epoch, trainloader, model, optimizer, loss_mse):
    model.train()
    train_loss = 0
    train_losses = []
    for batch_idx, data in enumerate(trainloader):
        data = data.to(device)
        optimizer.zero_grad()
        recon_batch, mu, logvar = model(data)
        loss = loss_mse(recon_batch, data, mu, logvar)
        loss.backward()
        train_loss += loss.item()
        optimizer.step()
    #        if batch_idx % log_interval == 0:
    #            print('Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}'.format(
    #                epoch, batch_idx * len(data), len(trainloader.dataset),
    #                       100. * batch_idx / len(trainloader),
    #                       loss.item() / len(data)))
    if epoch % 200 == 0:
        print(
            "====> Epoch: {} Average loss: {:.4f}".format(
                epoch, train_loss / len(trainloader.dataset)
            )
        )
        train_losses.append(train_loss / len(trainloader.dataset))
    return model, train_losses


# %%
# training
class VAE:
    def __init__(
        self,
        n_components=None,
        hidden_layers_dim=[64, 32],  # INFO: This argument can mutate.
        lr=1e-3,
        epochs=1500,
    ):
        self.n_components = n_components
        self.hidden_layers_dim = hidden_layers_dim
        self.lr = lr
        self.epochs = epochs

    def fit(self, x):
        data_set = DataBuilder(x)
        self.trainloader = DataLoader(dataset=data_set, batch_size=32)
        self.d_in = data_set.x.shape[1]

        if self.n_components is None:
            self.d_out = self.d_in
        else:
            self.d_out = self.n_components

        self.model = Autoencoder(
            self.d_in, h=self.hidden_layers_dim, latent_dim=self.d_out
        ).to(device)
        self.model.apply(weights_init_uniform_rule)
        self.optimizer = optim.Adam(self.model.parameters(), lr=self.lr)
        self.loss_mse = custom_loss()
        # self.epochs = self.epochs # INFO: This is likely a bug.
        # log_interval = 50
        # val_losses = []
        # train_losses = []
        self.__train()

    def __train(self):
        self.model, self.train_losses = train(
            self.epochs, self.trainloader, self.model, self.optimizer, self.loss_mse
        )

    def transform(self, x):
        data_set = DataBuilder(x)
        trainloader = DataLoader(dataset=data_set, batch_size=1024)
        # evaluate
        self.model.eval()
        # output space
        mu_output = []
        logvar_output = []

        with torch.no_grad():
            for i, (data) in enumerate(trainloader):
                data = data.to(device)
                recon_batch, mu, logvar = self.model(data)
                mu_tensor = mu
                mu_output.append(mu_tensor)
                mu_result = torch.cat(mu_output, dim=0)
                logvar_tensor = logvar
                logvar_output.append(logvar_tensor)
                self.logvar_result = torch.cat(logvar_output, dim=0)
        return mu_result.cpu().detach().numpy()


# #%%
# from mpl_toolkits import mplot3d


# import numpy as np
# import matplotlib.pyplot as plt
# #ax = plt.axes(projection='3d')
# #ax = plt.subplot()

# # Data for three-dimensional scattered points
# winetype = data_set.wine
# #zdata = mu_result[:,0].cpu().numpy()
# xdata = mu_result[:,1].cpu().numpy()
# ydata = mu_result[:,2].cpu().numpy()
# plt.scatter(xdata, ydata,c=df.idCluster )
# plt.show()

# #%%
# import umap.umap_ as umap
# nD = 3
# umapReducer = umap.UMAP(random_state=42000, n_components=nD)
# x = umapReducer.fit_transform( mu_result.cpu().numpy())
# #df_x2D = pd.DataFrame(x, columns=["D1", "D2", "D3"])
# #df_x2D = pd.concat([df['filename'], df_x2D, df[list(df_conf.columns[
# 1:])]], axis=1)

# # plt.scatter(x[:,0], x[:,1], c=df.idCluster)
# # plt.show()
# # ""


# import matplotlib.pyplot as plt
# ax = plt.axes(projection='3d')
# #ax = plt.subplot()
# # Data for three-dimensional scattered points
# winetype = data_set.wine
# zdata = mu_result[:,0].cpu().numpy()
# xdata = mu_result[:,1].cpu().numpy()
# ydata = mu_result[:,2].cpu().numpy()
# ax.scatter(xdata, ydata, zdata, c=df.idCluster )
# plt.show()



from processing.config.files.FileConfig import FileConfig
from processing.loaders.Loader import Loader


class FileLoader:
    def __init__(
        self,
        file: FileConfig,
    ) -> None:
        self.file = file
        self.loader = Loader()

    def load(self):
        self.loader.load(self.file)

    def release(self):
        self.loader.release()

    def destroy(self):
        self.loader.destroy()



from typing import Any, Dict, List, Optional, Union

import numpy
from h5py import Dataset, File
from rich import print

from processing.errors.DatasetTypeError import DatasetTypeError
from processing.storage.StorageCompression import StorageCompression
from processing.storage.StorageMode import StorageMode
from processing.storage.StoragePath import StoragePath
from processing.utils.get_version_from_setup import get_version_from_setup


class Storage:
    """The interface for handling HDF5 storage file."""

    def __init__(
        self,
        path: str,
    ) -> None:
        self.path = path
        self.__set_file_or_fail()

    def __set_file_or_fail(self) -> None:
        try:
            self.__file = File(
                self.path,
                StorageMode.rw_or_create.value,
            )
            print("success")
        except BlockingIOError:
            print("blocking error")
            raise RuntimeError(f"Unable to load file {self.path}.")
        except TypeError:
            print("type error")
            raise FileNotFoundError(f"Unable to find file {self.path}.")
        except OSError:
            raise TypeError(f"Unable to load file {self.path}.")

    def close(self) -> None:
        self.__file.close()

    def load(self) -> None:
        self.__set_file_or_fail()

    @staticmethod
    def __get_path_as_string(path: Union[StoragePath, str]) -> str:
        if type(path) is StoragePath:
            path = path.value

        path = str(path)
        return path

    def __write_version_to_dataset(
        self,
        dataset: Dataset,
    ) -> Dataset:
        # Attach app version to all datasets
        dataset.attrs["version"] = get_version_from_setup()
        return dataset

    def write(
        self,
        path: Union[StoragePath, str],
        data: Any,
        compression: Optional[bool] = False,
        dtype: Optional[Any] = None,
        shape: Optional[Any] = None,
        attributes: Optional[Dict[str, str]] = None,
    ) -> None:
        path = self.__get_path_as_string(path)

        if self.exists_dataset(path):
            print(f"Dataset already exists: {path}")
            return

        dataset = self.__file.create_dataset(
            path,
            data=data,
            compression=StorageCompression.gzip.value if compression else None,
            dtype=dtype,
            shape=shape,
        )

        if attributes is not None:
            for k, v in attributes.items():
                dataset.attrs[k] = v

        self.__write_version_to_dataset(dataset)

    def create_attribute(
        self,
        key: str,
        value: Any,
        path: Union[StoragePath, str],
    ) -> None:
        path = self.__get_path_as_string(path)

        if self.exists_attribute(path, key):
            print(f"Attribute already exists: {path} => {key}")
            return

        self.__file[path].attrs[key] = value

    def exists_attribute(
        self,
        path: Union[StoragePath, str],
        key: str,
    ) -> bool:
        path = self.__get_path_as_string(path)

        try:
            _ = self.read(path).attrs[key]
            return True
        except KeyError:
            return False

    def exists_dataset(
        self,
        path: Union[StoragePath, str],
    ) -> bool:
        path = self.__get_path_as_string(path)

        try:
            _ = self.read(path)
            return True
        except KeyError:
            return False

    # TODO: This can output a group!
    def read(
        self,
        path: Union[StoragePath, str],
    ) -> Dataset:
        try:
            path = self.__get_path_as_string(path)
            payload = self.__file[path]
            return payload  # type: ignore TODO
        except KeyError:
            raise KeyError(f"Unable to find storage path {path}.")

    # Silent delete
    def delete(
        self,
        path: Union[StoragePath, str],
    ) -> None:
        try:
            if isinstance(path, str):
                del self.__file[path]
            elif type(path) is StoragePath:
                del self.__file[path.value]
        except KeyError:
            return

    def write_empty_group(self, path: str) -> None:
        self.__file.create_group(path)

    def write_binary(
        self,
        path: str,
        binary_data: bytes,
    ) -> None:
        binary_array = numpy.frombuffer(binary_data, dtype="uint8")

        dataset = self.__file.create_dataset(
            name=path,
            data=binary_array,
            compression=StorageCompression.gzip.value,
        )

        self.__write_version_to_dataset(dataset)

    def append(
        self,
        path: str,
        data: Union[List[List[float]], List[List[str]]],
        compression: bool = False,
        attributes: Optional[Dict[str, str]] = None,
    ) -> None:
        length = len(data)
        dimensions = len(data[0])

        if not self.exists_dataset(path):
            dataset = self.__file.create_dataset(
                name=path,
                data=data,
                compression=(
                    StorageCompression.gzip.value if compression is True else None
                ),
                chunks=True,
                shape=(length, dimensions),
                maxshape=(None, dimensions),
            )

        else:
            dataset: Dataset = self.__file[path]  # type: ignore
            new_shape = dataset.shape[0] + length
            dataset.resize(new_shape, axis=0)
            dataset[-length:] = data

        if attributes is not None:
            for k, v in attributes.items():
                dataset.attrs[k] = v

        self.__write_version_to_dataset(dataset)

    @staticmethod
    def make_rectangular(
        non_rectangular_array: List[List[Any]],
        fill_with=None,
    ):
        # Determine the maximum length of the sub-arrays
        max_length = max(len(sub_array) for sub_array in non_rectangular_array)

        # Iterate through each sub-array and append None values to the end
        # until it has the maximum length
        rectangular_array = []
        for sub_array in non_rectangular_array:
            if len(sub_array) < max_length:
                sub_array += [fill_with] * (max_length - len(sub_array))
            rectangular_array.append(sub_array)

        # Return the new rectangular array
        return rectangular_array

    @staticmethod
    def trim_rectangular(
        rectangular_array,
        trim_with=None,
    ):
        jagged_array = []

        for sub_list in rectangular_array:
            trimmed_list = [e for e in sub_list if e is not trim_with]
            jagged_array.append(trimmed_list)

        return jagged_array

    @staticmethod
    def convert_dataset_to_string_list(dataset: Dataset) -> List[str]:
        (length,) = dataset.shape

        if length == 0:
            return []

        if dataset.dtype == "object":  # strings
            string_list = list(dataset.asstr()[:])
        elif dataset.dtype == "int64":  # numbers
            string_list = [str(v) for v in dataset]
        else:
            raise DatasetTypeError

        return string_list



from processing.utils.print_action import print_action


def get_file_full_path(filename: str, audio_folder: str):
    if filename[0] == "/":
        path = f"{audio_folder}{filename}"
    else:
        print_action(f"File name {filename} does not start with a /", "error")
        path = f"{audio_folder}/{filename}"

    return path



from typing import List

from processing.common.TrajectoryConfigError import TrajectoryConfigError
from processing.config.labels.LabelConfig import LabelConfig


def validate_trajectory_labels(
    names: List[str],
    properties: List[str],
    values: List[str],
    labels: List[LabelConfig],
) -> None:
    files_properties, files_values = LabelConfig.flatten(labels)

    for i, property in enumerate(properties):
        if property not in files_properties:
            raise TrajectoryConfigError(
                f"Property {property} not found in files for trajectory {names[i]}."
            )

        property_index = files_properties.index(property)
        value = values[i]
        files_value = files_values[property_index]

        if value not in files_value:
            raise TrajectoryConfigError(
                f"Value {value} not found in files for property"
                f" {property} for trajectory {names[i]}."
            )



from typing import Iterable, List, Tuple

from processing.config.bands.BandConfig import BandConfig
from processing.config.integrations.IntegrationConfig import IntegrationConfig


def walk_bands_integrations(
    bands: List[BandConfig],
    integrations: List[IntegrationConfig],
) -> Iterable[Tuple[BandConfig, IntegrationConfig]]:
    for band in bands:
        for integration in integrations:
            yield band, integration



import numpy as np

from processing.digesters.Digester import Digester


class SumVarianceDigester(Digester):
    def digest(self, labels):
        label = labels[0]
        data = []

        for _, frame, _ in self.walk_within_label(label):
            var = np.var(frame, axis=0)
            sum_var = np.sum(var)
            data.append(float(sum_var))

        return data



from processing.config.digesters.DigesterStorage import DigesterStorage
from processing.interfaces import MenuCallback
from processing.storage.Storage import Storage
from processing.utils.invoke_menu import invoke_menu


def validate_digesters(action):
    """The digesters configuration objects validator."""

    def decorator(
        storage: Storage,
        callback: MenuCallback,
    ):
        digesters = DigesterStorage.read_from_storage(storage)

        if len(digesters) == 0:
            invoke_menu(storage, callback)
            return

        return action(storage, callback)

    return decorator



import os
import platform
from typing import List


def walk_directory(path: str) -> List[str]:
    relative_paths = []
    supported_formats = [".wav", ".mp3", ".flac"]

    for dir_path, _, filenames in os.walk(path):
        for f in filenames:
            for format_ in supported_formats:
                if f.endswith(format_):
                    absolute_path = os.path.join(dir_path, f)
                    relative_path = absolute_path.replace(path, "")

                    if platform.system() == "Windows":
                        relative_path = relative_path.replace("\\", "/")

                    relative_paths.append(relative_path)

    return relative_paths



from typing import List

import pandas
from torch.utils.data import Dataset

from processing.common.ContinuousTimeTrajectory import ContinuousTimeTrajectory
from processing.config.trajectories.TrajectoryConfig import TrajectoryConfig


def read_trajectory_path_and_relative_timestamps(
    trajectory: TrajectoryConfig,
    features: Dataset,
    timestamps: Dataset,
    labels_properties: List[str],
    labels_values: Dataset,
    label_property: str,
    label_value: str,
):
    ctt = ContinuousTimeTrajectory()
    ctt.load(
        features=features,
        timestamps=timestamps,
        timestamp_start=trajectory.start,
        timestamp_end=trajectory.end,
        labels_properties=labels_properties,
        labels_values=labels_values,
        step=trajectory.step,
    )

    ctt.calculate(
        trajectory_label_property=label_property,
        trajectory_label_value=label_value,
    )

    path = pandas.DataFrame(ctt.values)
    relative_timestamps = pandas.DataFrame(ctt.relative_timestamps)

    return path, relative_timestamps



from typing import Optional

from InquirerPy.resolver import prompt
from InquirerPy.separator import Separator

from processing.common.MenuChoice import MenuChoice
from processing.storage.Storage import Storage
from processing.utils.print_menu_legend import print_menu_legend
from processing.utils.wrap_menu_choice_with_state import wrap_menu_choice_with_state


def ask_menu(
    storage: Storage,
    last_choice: Optional[MenuChoice] = None,
) -> MenuChoice:
    print_menu_legend()

    questions = [
        {
            "type": "list",
            "name": "choices",
            "choices": [
                MenuChoice.RunAll.value,
                Separator(),
                wrap_menu_choice_with_state(MenuChoice.RefreshConfig, storage),
                wrap_menu_choice_with_state(MenuChoice.ExtractAggregate, storage),
                wrap_menu_choice_with_state(MenuChoice.Reduce, storage),
                Separator(),
                wrap_menu_choice_with_state(MenuChoice.ComputeRequirements, storage),
                MenuChoice.PurgeRequirements.value,
                Separator(),
                wrap_menu_choice_with_state(MenuChoice.Autocluster, storage),
                wrap_menu_choice_with_state(MenuChoice.Trace, storage),
                wrap_menu_choice_with_state(MenuChoice.RelativeTrace, storage),
                wrap_menu_choice_with_state(MenuChoice.Digest, storage),
                Separator(),
                MenuChoice.ExportDataframe.value,
                MenuChoice.ExportComputationUmaps.value,
                MenuChoice.ExportMeanDistancesMatrix.value,
                Separator(),
                MenuChoice.Quit.value,
            ],
            "message": "Choose your action",
            "default": last_choice,
        }
    ]

    answers = prompt(questions=questions, vi_mode=True)
    answer: MenuChoice = str(answers["choices"])  # type: ignore
    return answer



from typing import List, Tuple

from h5py import Dataset


def filter_features_by_label_and_time(
    features: Dataset,
    timestamps: Dataset,
    timestamp_start: int,  # UNIX timestamp in milliseconds
    timestamp_end: int,  # UNIX timestamp in milliseconds
    labels_properties: List[str],
    labels_values: Dataset,
    trajectory_label_property: str,
    trajectory_label_value: str,
) -> Tuple[List[List[float]], List[int]]:
    filtered_features = []
    filtered_timestamps: List[int] = []
    label_property_index = labels_properties.index(trajectory_label_property)

    for index, timestamp in enumerate(timestamps):
        label_values = labels_values[index]
        label_value = label_values[label_property_index].decode("utf-8")

        if (
            timestamp[0] <= timestamp_start
            or timestamp[0] >= timestamp_end
            or label_value != trajectory_label_value
        ):
            continue

        filtered_features.append(features[index])
        filtered_timestamps.append(timestamp[0])

    return filtered_features, filtered_timestamps



from typing import List

from InquirerPy.resolver import prompt
from rich import print

from processing.config.integrations.IntegrationConfig import IntegrationConfig


def ask_integration(integrations: List[IntegrationConfig]) -> IntegrationConfig:
    questions = [
        {
            "type": "list",
            "name": "choices",
            "choices": [integration.name for integration in integrations],
            "message": "Choose your integration",
        }
    ]

    print()
    answers = prompt(questions)
    name: str = str(answers["choices"])

    results: List[IntegrationConfig] = list(
        filter(lambda i: i.name == name, integrations)
    )

    assert len(results) == 1, f"{len(results)} results for {name}"

    integration = results[0]

    return integration



from typing import Literal

from rich import print


def print_action(
    payload: str,
    mode: Literal["start", "end", "error", "warning"],
) -> None:
    print()

    if mode == "start":
        print(f"[bold yellow]:rocket: {payload}[/bold yellow]")
    if mode == "end":
        print(f"[bold green]:party_popper: {payload}[/bold green]")
    if mode == "error":
        print(f"[bold red]:collision: {payload}[/bold red]")
    if mode == "warning":
        print(f"[yellow]:warning: {payload}[/yellow]")

    print()



from abc import ABC, abstractmethod
from typing import Any, Iterable, List, Optional

from processing.config.bands.BandConfig import BandConfig
from processing.loaders.Loader import Loader
from processing.loaders.SoundSlice import SoundSlice
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath

Extracted = List[List[Any]]


class Extractor(ABC):
    index: int
    __expected_sample_rate: Optional[int] = None
    __band: Optional[BandConfig] = None
    __offset: Optional[int] = None
    __step: Optional[int] = None
    is_persist: bool = False

    @property
    def path(self):
        assert self.index is not None, "Please add an extractor index"
        return f"{StoragePath.extracted.value}/{self.band.name}/{self.index}"

    @property
    def expected_sample_rate(self) -> int:
        assert (
            self.__expected_sample_rate is not None
        ), "Please define expected sample rate"
        return self.__expected_sample_rate

    @expected_sample_rate.setter
    def expected_sample_rate(self, expected_sample_rate: int) -> int:
        self.__expected_sample_rate = expected_sample_rate
        return self.__expected_sample_rate

    @property
    def band(self) -> BandConfig:
        assert self.__band is not None, "Please define band"
        return self.__band

    @band.setter
    def band(self, band: BandConfig) -> BandConfig:
        self.__band = band
        return self.__band

    @property
    def offset(self) -> int:
        assert self.__offset is not None, "Please define offset"
        return self.__offset

    @offset.setter
    def offset(self, offset: int) -> int:
        self.__offset = offset
        return self.__offset

    @property
    def step(self) -> int:
        assert self.__step is not None, "Please define step"
        return self.__step

    @step.setter
    def step(self, step: int) -> int:
        self.__step = step
        return self.__step

    @abstractmethod
    def extract(self, loader: Loader) -> Extracted:
        pass

    def persist(self):
        self.is_persist = True

    def store(self, data: Extracted, storage: Storage):
        storage.append(
            path=self.path,
            data=data,
            compression=True,
            attributes={
                "extractor": self.__class__.__name__,
                "offset": str(self.offset),
                "step": str(self.step),
                "method": "for step in file in site",
            },
        )

    def sound_walk(self, loader: Loader) -> Iterable[SoundSlice]:
        sample_rate = loader.sound.sample_rate
        filtered = loader.sound.get_filtered(self.band)

        t = int(self.offset / 1000 * sample_rate)
        step = int(self.step / 1000 * sample_rate)

        while t < len(filtered):
            start = t
            end = start + step

            slice_ = filtered[start:end]

            # skipping for incomplete step
            if len(slice_) < step:
                t += step
                continue

            t += step

            yield SoundSlice(slice_, start, end)



from processing.extractors.Extractor import Extractor
from processing.loaders.Loader import Loader


class TemporalMedianExtractor(Extractor):
    """med"""

    def extract(self, loader: Loader):
        from maad import features

        data = []

        for slice_ in self.sound_walk(loader):
            temporal_median = features.temporal_median(slice_.sound)
            data.append([temporal_median])

        return data



from InquirerPy.resolver import prompt
from rich import print


def ask_for_repack_replacement() -> bool:
    questions = [
        {
            "type": "confirm",
            "message": "Do you want to replace your old storage file?",
            "name": "replace",
            "default": False,
        },
    ]

    print()
    answers = prompt(questions)
    answer: bool = bool(answers["replace"])
    return answer



from typing import Any, List


def get_uniques_unsorted(array: List[Any]) -> List[str]:
    uniques = []

    for element in array:
        string = str(element)

        if string in uniques:
            continue

        uniques.append(string)

    return uniques



from processing.constants import DOCKER_BASE_PATH


def read_audio_path_in_docker(audio_path: str) -> str:
    base_path = DOCKER_BASE_PATH
    audio_folder = audio_path.split("/")[-1]
    docker_path = f"{base_path}/{audio_folder}"
    return docker_path



from typing import Dict, List, Tuple, Type

from processing.config.bands.BandConfig import BandConfig
from processing.config.integrations.IntegrationConfig import IntegrationConfig
from processing.config.ranges.RangeConfig import RangeConfig
from processing.reducers.AbstractReducer import AbstractReducer
from processing.reducers.PcaReducer import PcaReducer
from processing.reducers.SparsePcaReducer import SparsePcaReducer
from processing.reducers.UmapReducer import UmapReducer
from processing.reducers.VaeReducer import VaeReducer
from processing.utils.is_nan import is_nan


class ReducerConfig:
    algorithms: Dict[str, Type[AbstractReducer]] = {
        "umap": UmapReducer,
        "vae": VaeReducer,
        "pca": PcaReducer,
        "sparse_pca": SparsePcaReducer,
    }

    index: int
    name: str
    dimensions: int
    bands: List[BandConfig]
    integrations: List[IntegrationConfig]
    ranges: List[RangeConfig]
    band: BandConfig
    integration: IntegrationConfig
    instance: AbstractReducer

    def __init__(
        self,
        index: int,
        name: str,
        dimensions: int,
        bands: List[BandConfig],
        integrations: List[IntegrationConfig],
        ranges: List[RangeConfig],
    ) -> None:
        self._validate_name(name)

        self.index = index
        self.name = name
        self.dimensions = dimensions

        self.bands = bands
        self.integrations = integrations
        self.ranges = ranges

    @staticmethod
    def _validate_name(name: str) -> None:
        if name in ReducerConfig.algorithms.keys():
            return

        raise KeyError(f"Unable to find reducer name {name}.")

    def should_calculate(
        self,
    ) -> bool:
        bands_names = [b.name for b in self.bands]
        integrations_names = [i.name for i in self.integrations]

        if (
            self.band.name in bands_names
            and self.integration.name in integrations_names
        ):
            return True

        return False

    @staticmethod
    def flatten(
        reducers: List["ReducerConfig"],
    ) -> Tuple[
        List[str],
        List[int],
        List[List[str]],
        List[List[str]],
        List[List[str]],
    ]:
        names = [r.name for r in reducers]
        dimensions = [r.dimensions for r in reducers]
        bands_names = [[b.name for b in r.bands] for r in reducers]
        integrations_names = [[i.name for i in r.integrations] for r in reducers]
        ranges_names = [[r.name for r in r.ranges] for r in reducers]

        return names, dimensions, bands_names, integrations_names, ranges_names

    @staticmethod
    def pick_reducer_bands(
        bands_names_string: str,
        bands: List[BandConfig],
    ) -> List[BandConfig]:
        if is_nan(bands_names_string):
            return bands

        bands_names = bands_names_string.split(",")

        picked_bands = [band for band in bands if band.name in bands_names]

        return picked_bands

    @staticmethod
    def pick_reducer_integrations(
        integrations_names_string: str,
        integrations: List[IntegrationConfig],
    ) -> List[IntegrationConfig]:
        if is_nan(integrations_names_string):
            return integrations

        integrations_names = integrations_names_string.split(",")

        picked_integrations = [
            integration
            for integration in integrations
            if integration.name in integrations_names
        ]

        return picked_integrations

    @staticmethod
    def pick_reducer_ranges(
        ranges_names_string: str,
        ranges: List[RangeConfig],
    ) -> List[RangeConfig]:
        if is_nan(ranges_names_string):
            return ranges

        ranges_names = ranges_names_string.split(",")

        picked_ranges = [range for range in ranges if range.name in ranges_names]

        return picked_ranges

    @staticmethod
    def reconstruct(
        names: List[str],
        dimensions: List[int],
        bands_names_strings: List[str],
        integrations_names_strings: List[str],
        ranges_names_strings: List[str],
        bands: List[BandConfig],
        integrations: List[IntegrationConfig],
        ranges: List[RangeConfig],
    ) -> List["ReducerConfig"]:
        reducers = []

        for index, name in enumerate(names):
            reducer_bands = ReducerConfig.pick_reducer_bands(
                bands_names_string=bands_names_strings[index],
                bands=bands,
            )

            reducer_integrations = ReducerConfig.pick_reducer_integrations(
                integrations_names_string=integrations_names_strings[index],
                integrations=integrations,
            )

            reducer_ranges = ReducerConfig.pick_reducer_ranges(
                ranges_names_string=ranges_names_strings[index],
                ranges=ranges,
            )

            reducer = ReducerConfig(
                index=index,
                name=name,
                dimensions=dimensions[index],
                bands=reducer_bands,
                integrations=reducer_integrations,
                ranges=reducer_ranges,
            )

            reducers.append(reducer)

        return reducers

    def load(
        self,
        band: BandConfig,
        integration: IntegrationConfig,
    ):
        self.band = band
        self.integration = integration
        return self

    def create_instance(
        self,
        band: BandConfig,
        integration: IntegrationConfig,
    ) -> AbstractReducer:
        self.load(band=band, integration=integration)
        self.instance = self.algorithms[self.name]()
        return self.instance



from processing.extractors.Extractor import Extractor
from processing.loaders.Loader import Loader


class AcousticDiversityIndexExtractor(Extractor):
    """adi"""

    def extract(self, loader: Loader):
        from maad import features

        data = []

        for slice_ in self.sound_walk(loader):
            spectrogram = loader.sound.get_spectrogram(slice_, "amplitude")

            adi = features.acoustic_diversity_index(
                Sxx=spectrogram.s,
                fn=spectrogram.fn,
                fmax=10000,
                dB_threshold=-30,
            )

            data.append([adi])

        return data



from typing import List

from pandas import DataFrame


def convert_dataframe_to_list(
    df: DataFrame,
    flatten: bool = True,
) -> List[float]:
    list_ = df.to_numpy()

    if flatten is True:
        list_.flatten()

    return list(list_)



import os
from typing import List

from rich import print
from rich.console import Console
from rich.table import Table

from processing.config.autoclusters.AutoclusterConfig import AutoclusterConfig
from processing.config.autoclusters.AutoclusterStorage import AutoclusterStorage
from processing.config.bands.BandConfig import BandConfig
from processing.config.bands.BandStorage import BandStorage
from processing.config.binary.BinaryStorage import BinaryStorage
from processing.config.ConfigParser import ConfigParser
from processing.config.digesters.DigesterConfig import DigesterConfig
from processing.config.digesters.DigesterStorage import DigesterStorage
from processing.config.extractors.ExtractorConfig import ExtractorConfig
from processing.config.extractors.ExtractorStorage import ExtractorStorage
from processing.config.files.FileConfig import FileConfig
from processing.config.files.FileStorage import FileStorage
from processing.config.integrations.IntegrationConfig import IntegrationConfig
from processing.config.integrations.IntegrationStorage import IntegrationStorage
from processing.config.labels.LabelConfig import LabelConfig
from processing.config.labels.LabelStorage import LabelStorage
from processing.config.ranges.RangeConfig import RangeConfig
from processing.config.ranges.RangeStorage import RangeStorage
from processing.config.reducers.ReducerConfig import ReducerConfig
from processing.config.reducers.ReducerStorage import ReducerStorage
from processing.config.settings.SettingsConfig import SettingsConfig
from processing.config.settings.SettingsRow import SettingsRow
from processing.config.settings.SettingsStorage import SettingsStorage
from processing.config.sites.SiteConfig import SiteConfig
from processing.config.sites.SiteStorage import SiteStorage
from processing.config.trajectories.TrajectoryConfig import TrajectoryConfig
from processing.config.trajectories.TrajectoryStorage import TrajectoryStorage
from processing.storage.Storage import Storage
from processing.utils.convert_timestamp_to_date import convert_timestamp_to_date


class Config:
    def __init__(
        self,
        path: str,
    ) -> None:
        self.__print_load()
        self.path = os.path.abspath(path)
        self.folder = os.path.dirname(self.path)
        self.parser = ConfigParser(self.path, self.folder)

        self.settings: SettingsConfig

        self.bands: List[BandConfig] = []
        self.integrations: List[IntegrationConfig] = []
        self.ranges: List[RangeConfig] = []

        self.labels: List[LabelConfig] = []
        self.files: List[FileConfig] = []
        self.sites: List[SiteConfig] = []
        self.extractors: List[ExtractorConfig] = []

        self.autoclusters: List[AutoclusterConfig] = []
        self.trajectories: List[TrajectoryConfig] = []

        self.reducers: List[ReducerConfig] = []
        self.digesters: List[DigesterConfig] = []

        self.parse()
        self.__print_success()

    def __print_load(self) -> None:
        print("Loading configuration...")

    def __print_success(self) -> None:
        print(f"Config loaded: {self.parser.path}")
        self.print_settings()

    def print_settings(self) -> None:
        console = Console()

        table = Table(show_header=True, header_style="bold magenta")
        table.add_column("Setting")
        table.add_column("Value")

        for k, v in vars(self.settings).items():
            if k == SettingsRow.timeline_origin.value:
                v = convert_timestamp_to_date(v)

            table.add_row(str(k), str(v))

        console.print(table)

    def parse(self) -> None:
        self.settings = SettingsStorage.read_from_config(self.parser)

        self.bands = BandStorage.read_from_config(self.parser)
        self.integrations = IntegrationStorage.read_from_config(self.parser)
        self.ranges = RangeStorage.read_from_config(self.parser)

        self.labels = LabelStorage.read_from_config(self.parser)
        self.files = FileStorage.read_from_config(
            parser=self.parser,
            labels=self.labels,
            settings=self.settings,
        )

        self.sites = SiteStorage.parse_from_config(self.files)
        self.extractors = ExtractorStorage.read_from_config(self.parser)

        self.autoclusters = AutoclusterStorage.read_from_config(self.parser)
        self.trajectories = TrajectoryStorage.read_from_config(
            parser=self.parser,
            labels=self.labels,
        )

        self.reducers = ReducerStorage.read_from_config(
            parser=self.parser,
            bands=self.bands,
            integrations=self.integrations,
            ranges=self.ranges,
        )

        self.digesters = DigesterStorage.read_from_config(self.parser)

    def delete_from_storage(self, storage: Storage) -> None:
        BinaryStorage.delete_from_storage(storage)
        SettingsStorage.delete_from_storage(storage)

        BandStorage.delete_from_storage(storage)
        IntegrationStorage.delete_from_storage(storage)
        RangeStorage.delete_from_storage(storage)

        LabelStorage.delete_from_storage(storage)
        FileStorage.delete_from_storage(storage)
        SiteStorage.delete_from_storage(storage)

        ExtractorStorage.delete_from_storage(storage)

        AutoclusterStorage.delete_from_storage(storage)
        TrajectoryStorage.delete_from_storage(storage)

        ReducerStorage.delete_from_storage(storage)
        DigesterStorage.delete_from_storage(storage)

    def write(self, storage: Storage) -> None:
        self.delete_from_storage(storage)

        BinaryStorage.write_to_storage(self.path, storage)
        SettingsStorage.write_to_storage(self.settings, storage)

        BandStorage.write_to_storage(self.bands, storage)
        IntegrationStorage.write_to_storage(self.integrations, storage)
        RangeStorage.write_to_storage(self.ranges, storage)

        LabelStorage.write_to_storage(self.labels, storage)
        FileStorage.write_to_storage(self.files, storage)
        SiteStorage.write_to_storage(self.sites, storage)

        ExtractorStorage.write_to_storage(self.extractors, storage)

        AutoclusterStorage.write_to_storage(self.autoclusters, storage)
        TrajectoryStorage.write_to_storage(self.trajectories, storage)

        ReducerStorage.write_to_storage(self.reducers, storage)
        DigesterStorage.write_to_storage(self.digesters, storage)

    @staticmethod
    def exists_in_storage(storage: Storage) -> bool:
        return (
            BinaryStorage.exists_in_storage(storage)
            and SettingsStorage.exists_in_storage(storage)
            and BandStorage.exists_in_storage(storage)
            and IntegrationStorage.exists_in_storage(storage)
            and RangeStorage.exists_in_storage(storage)
            and LabelStorage.exists_in_storage(storage)
            and FileStorage.exists_in_storage(storage)
            and SiteStorage.exists_in_storage(storage)
            and ExtractorStorage.exists_in_storage(storage)
            and AutoclusterStorage.exists_in_storage(storage)
            and TrajectoryStorage.exists_in_storage(storage)
            and ReducerStorage.exists_in_storage(storage)
            and DigesterStorage.exists_in_storage(storage)
        )



from typing import List

from processing.config.ConfigParser import ConfigParser
from processing.config.ExcelSheet import ExcelSheet
from processing.config.files.FileConfig import FileConfig
from processing.config.files.FileSheet import FileSheet
from processing.config.labels.LabelConfig import LabelConfig
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath


class LabelStorage:
    properties = StoragePath.labels_properties.value
    sets = StoragePath.labels_sets.value

    @staticmethod
    def delete_from_storage(storage: Storage) -> None:
        storage.delete(LabelStorage.properties)
        storage.delete(LabelStorage.sets)

    @staticmethod
    def exists_in_storage(storage: Storage) -> bool:
        return storage.exists_dataset(
            LabelStorage.properties
        ) and storage.exists_dataset(LabelStorage.sets)

    @staticmethod
    def read_from_storage(
        storage: Storage,
        files: List[FileConfig],
    ) -> List[LabelConfig]:
        properties_dataset = storage.read(LabelStorage.properties)
        properties = storage.convert_dataset_to_string_list(properties_dataset)

        labels: List[LabelConfig] = []

        for index, property in enumerate(properties):
            label: LabelConfig = LabelConfig(index, property)
            values = [file.labels[index] for file in files]
            label.load_values(values)
            labels.append(label)

        return labels

    @staticmethod
    def read_properties_from_storage(
        storage: Storage,
    ) -> List[str]:
        properties_dataset = storage.read(LabelStorage.properties)
        properties = storage.convert_dataset_to_string_list(properties_dataset)
        return properties

    @staticmethod
    def read_from_config(parser: ConfigParser) -> List[LabelConfig]:
        sheet = parser.parse(ExcelSheet.files)

        labels: List[LabelConfig] = []

        index = 0
        for column in sheet:
            if not LabelConfig.is_label_property(column):
                continue

            property_ = LabelConfig.trim_prefixed_property_from_config(column)
            label: LabelConfig = LabelConfig(index=index, property=property_)

            values = parser.get(
                ExcelSheet.files,
                FileSheet.label_prefix,
                suffix=label.property,
            )

            label.load_values(values)
            labels.append(label)
            index += 1

        return labels

    @staticmethod
    def write_to_storage(labels: List[LabelConfig], storage: Storage) -> None:
        if len(labels) == 0:
            properties = []
            sets_rectangular = []
        else:
            properties, sets = LabelConfig.flatten(labels)
            sets_rectangular = storage.make_rectangular(sets, "")

        storage.write(path=LabelStorage.properties, data=properties)
        storage.write(path=LabelStorage.sets, data=sets_rectangular)



import io

from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath


def extract_config_from_storage(storage_path: str):
    storage = Storage(storage_path)
    binary_dataset = storage.read(StoragePath.config_file.value)
    version = binary_dataset.attrs["version"]
    binary_data: bytes = binary_dataset[:].tobytes()

    file_object = io.BytesIO(binary_data)
    file_path = f"{storage_path}-config-{version}.xlsx"

    with open(file_path, "wb") as f:
        f.write(file_object.getbuffer())



from typing import List, Union

from mutagen import File  # type: ignore
from mutagen.flac import FLAC
from mutagen.mp3 import MP3
from mutagen.wave import WAVE

from processing.utils.get_file_full_path import get_file_full_path

MyFile = Union[WAVE, MP3, FLAC, None]


def read_files_durations(
    names: List[str],
    base_path: str,
) -> List[int]:
    durations: List[int] = []

    for name in names:
        path = get_file_full_path(name, base_path)
        file: MyFile = File(path, easy=True)

        assert file is not None, f"File {path} not found"

        length = file.info.length

        assert (
            type(length) is int or type(length) is float
        ), f"Invalid file length type {type(file.info.length)}"

        milliseconds = int(length * 1000)
        durations.append(milliseconds)

    return durations



from typing import Dict, List, Type

from processing.digesters.ContingencyDigester import ContingencyDigester
from processing.digesters.Digester import Digester
from processing.digesters.DistanceDigester import DistanceDigester
from processing.digesters.MeanSpreadingDigester import MeanSpreadingDigester
from processing.digesters.MeanStandardDeviationDigester import (
    MeanStandardDeviationDigester,
)
from processing.digesters.OverlapDigester import OverlapDigester
from processing.digesters.SilhouetteDigester import SilhouetteDigester
from processing.digesters.SumStandardDeviationDigester import (
    SumStandardDeviationDigester,
)
from processing.digesters.SumVarianceDigester import SumVarianceDigester


class DigesterConfig:
    digesters: Dict[str, Type[Digester]] = {
        "sum_var": SumVarianceDigester,
        "sum_std": SumStandardDeviationDigester,
        "mean_std": MeanStandardDeviationDigester,
        "mean_spreading": MeanSpreadingDigester,
        "distance": DistanceDigester,
        "overlap": OverlapDigester,
        "silhouette": SilhouetteDigester,
        "contingency": ContingencyDigester,
    }

    pairings = ["contingency"]

    def __init__(
        self,
        index: int,
        name: str,
    ) -> None:
        self.validate_name(name=name)

        self.index = index
        self.name = name

        self.is_pairing = True if self.name in self.pairings else False

    def validate_name(self, name: str) -> None:
        keys = self.digesters.keys()
        assert name in keys, f"Unable to find digester name {name}"

    def instantiate(self) -> Digester:
        instance = self.digesters[self.name]()
        instance.index = self.index
        instance.is_pairing = self.is_pairing
        return instance

    @staticmethod
    def flatten(digesters: List["DigesterConfig"]) -> List[str]:
        names = [d.name for d in digesters]
        return names

    @staticmethod
    def reconstruct(names: List[str]) -> List["DigesterConfig"]:
        digesters = []

        for index, name in enumerate(names):
            digester = DigesterConfig(
                index=index,
                name=name,
            )

            digesters.append(digester)

        return digesters



from typing import List

from rich.console import Console
from rich.table import Table

from processing.common.AggregatedReduceable import AggregatedReduceable
from processing.storage.Storage import Storage


def print_aggregated_reduceables(
    ars: List[AggregatedReduceable],
    storage: Storage,
):
    console = Console()

    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("From extractor")
    table.add_column("Band")
    table.add_column("Integration")
    table.add_column("Data in storage")

    for ar in ars:
        table.add_row(
            ar.extractor.name,
            ar.band.name,
            str(ar.integration.seconds),
            "[green]:heavy_check_mark:[/green]"
            if ar.exists_in_storage(storage)
            else "[red]:x:[/red]",
        )

    console.print(table)
